from wq.io import TimeSeriesMapper, XmlParser, BaseIO
from datetime import datetime, date, timedelta
from climata.base import WebserviceLoader, FilterOpt, DateOpt, ChoiceOpt


class CocorahsIO(WebserviceLoader, XmlParser, TimeSeriesMapper, BaseIO):
    """
    Retrieves CoCoRaHS observations from data.cocorahs.org

    Usage:

    data = CocorahsIO(state='MN', county='HN')
    for row in data:
        print row.stationname, row.observationdate.date(), row.totalprecipamt
    """

    # Customize date parameters
    start_date = DateOpt(
        required=True,
        date_only=False,
        url_param="StartDate",
    )
    end_date = DateOpt(
        date_only=False,
        url_param="EndDate",
    )

    # These region filters are supported
    state = FilterOpt(required=True)
    county = FilterOpt()

    # Other filters are ignored
    basin = FilterOpt(ignored=True)
    station = FilterOpt(ignored=True)
    parameter = FilterOpt(ignored=True)

    # CoCoRaHS-specific options
    datetype = ChoiceOpt(
        url_param="ReportDateType",
        default="reportdate",
        choices=["reportdate", "timestamp"],
    )
    reporttype = ChoiceOpt(
        url_param="ReportType",
        default="Daily",
        choices=["Daily", "MultiDay"],
    )

    # Configuration for wq.io base classes
    url = "http://data.cocorahs.org/cocorahs/export/exportreports.aspx"

    root_tag = 'Cocorahs'

    date_formats = [
        '%Y-%m-%d %I:%M %p',
        '%Y-%m-%d',
        '%I:%M %p',
        '%m/%d/%Y %I:%M %p'
    ]

    key_fields = [
        "stationnumber",
        "stationname",
        "latitude",
        "longitude",
        "datetimestamp",
        "observationdate",
        "observationtime",
        "entrydatetime",
    ]

    # These params apply to every request
    default_params = {
        'dtf': "1",
        'Format': "XML",
        'TimesInGMT': "False",
        'responsefields': "all"
    }

    @property
    def item_tag(self):
        if self.getvalue('reporttype') == "Daily":
            return 'DailyPrecipReports/DailyPrecipReport'
        else:
            # i.e. self.getvalue('reporttype') == "MultiDay"
            return 'MultiDayPrecipReports/MultiDayPrecipReport'

    def serialize_params(self, params, complex):
        params = super(CocorahsIO, self).serialize_params(params, complex)
        fmt = '%m/%d/%Y'

        # Different date parameters and formats depending on use case
        if 'EndDate' in params:
            # Date range (usually used with datetype=reportdate)
            params['StartDate'] = self.getvalue('start_date').strftime(fmt)
            params['EndDate'] = self.getvalue('end_date').strftime(fmt)
        else:
            # Only start date (usually used with datetype=timestamp)
            params['Date'] = self.getvalue(
                'start_date'
            ).strftime(fmt + " %I:%M %p")
            del params['StartDate']
        return params

    def map_value(self, field, value):
        value = super(CocorahsIO, self).map_value(field, value)
        # CoCoRaHS empty dates are represented as 1/1/0001
        if isinstance(value, datetime) and value.year == 1:
            return None
        return value



from wq.io import BaseIO, TupleMapper, TimeSeriesMapper
from wq.io.parsers.base import BaseParser
from wq.io.exceptions import NoData
from wq.io.util import flattened

from SOAPpy import SOAPProxy, simplify
from climata.base import WebserviceLoader, FilterOpt, DateOpt, ChoiceOpt
from climata.base import fill_date_range, as_list

namespace = 'http://www.wcc.nrcs.usda.gov/ns/awdbWebService'
url = 'http://www.wcc.nrcs.usda.gov/awdbWebService/services?WSDL'
server = SOAPProxy(url, namespace)


class SnotelIO(WebserviceLoader, BaseParser, TupleMapper, BaseIO):
    """
    Base class for accessing SNOTEL AWDB SOAP web services.
    """

    webservice_name = "awdbWebService"
    data_function = None

    # Override Default WebserviceLoader options
    start_date = DateOpt(ignored=True)
    end_date = DateOpt(ignored=True)
    state = FilterOpt(ignored=True)
    basin = FilterOpt(ignored=True)
    county = FilterOpt(ignored=True)
    station = FilterOpt(ignored=True)
    parameter = FilterOpt(ignored=True)

    def load(self):
        if self.debug:
            self.print_debug()
        params = self.params
        fn = getattr(server, self.data_function)
        self.data = simplify(fn(**params))
        if self.data == {}:
            self.data = []
        else:
            self.data = as_list(self.data)

    def parse(self):
        # Some records may have additional fields; loop through entire
        # array to ensure all field names are accounted for.  (Otherwise BaseIO
        # will guess field names using only the first record.)

        field_names = set()
        for row in self.data:
            field_names.update(row.keys())
        self.field_names = field_names

    def print_debug(self):
        print '%s.%s(%s)' % (
            self.webservice_name,
            self.data_function,
            ','.join(
                '%s=%s' % (key, val)
                for key, val in self.params.items()
            )
        )


class StationIO(SnotelIO):
    """
    Retrieve metadata for all stations in a region.  Leverages both
    getStations() and getStationMetadata().
    """

    data_function = 'getStations'

    # Applicable WebserviceLoader default options
    state = FilterOpt(url_param='stateCds', multi=True)
    county = FilterOpt(url_param='countyNames', multi=True)
    basin = FilterOpt(url_param='hucs', multi=True)
    parameter = FilterOpt(url_param='elementCds', multi=True)

    # Additional options
    min_latitude = FilterOpt(url_param='minLatitude')
    max_latitude = FilterOpt(url_param='maxLatitude')
    min_elevation = FilterOpt(url_param='minElevation')
    max_elevation = FilterOpt(url_param='maxElevation')
    ordinals = FilterOpt(url_param='ordinals')

    # This is not the same as station (stationTriplet)
    station_ids = FilterOpt(url_param='stationIds', multi=True)

    # heightDepths = FilterOpt(url_param='heightDepths')
    # This parameter is submitted as
    # <heightDepths><value>value</value><unitCd>unit</unitCd></heightDepths>
    # Left out since it doesn't seem important and isn't well-documented

    default_params = {
        'logicalAnd': 'true',
    }

    def load(self):
        super(StationIO, self).load()
        self.data = [
            StationMetaIO(station=station, debug=self.debug).data[0]
            for station in self.data
        ]


class StationMetaIO(SnotelIO):
    """
    Wrapper for getStationMetadata() - used internally by StationIO.
    """

    data_function = 'getStationMetadata'

    station = FilterOpt(required=True, url_param='stationTriplet')


class StationElementIO(SnotelIO):
    """
    Wrapper for getStationElements(), incorporating element names from
    getElements()
    """

    data_function = 'getStationElements'

    # Applicable WebserviceLoader default options
    start_date = DateOpt(url_param='beginDate')
    end_date = DateOpt(url_param='endDate')
    station = FilterOpt(required=True, url_param='stationTriplet')

    def load(self):
        super(StationElementIO, self).load()
        names = {e.elementcd: e.name for e in elements}
        for elem in self.data:
            elem['element_name'] = names[elem['elementCd']]


class StationDataIO(StationElementIO):
    """
    Base class for StationDailyDataIO and StationHourlyDataIO.  Retrieves all
    data for a station that matches the specified duration by calling the
    specified inner_io_class.
    """
    nested = True

    # Applicable WebserviceLoader default options
    start_date = DateOpt(url_param='beginDate', required=True)
    end_date = DateOpt(url_param='endDate', required=True)
    parameter = FilterOpt()

    inner_io_class = None
    duration = None

    @property
    def params(self):
        params = super(StationDataIO, self).params
        # Parameter filter (if any) is applied *after* the initial request
        params.pop('parameter', None)
        return params

    def load(self):
        super(StationDataIO, self).load()
        data = []
        for row in self.data:

            # Only include records matching the specified duration
            # and parameter
            if row['duration'] != self.duration:
                continue
            elem = self.getvalue('parameter')
            if elem and row['elementCd'] != elem:
                continue

            # getStationElements() sometimes returns parameters that don't
            # actually have data for the requested timeframe - silently catch
            # the exception and remove parameter from results.
            try:
                row['data'] = self.inner_io_class(
                    station=row['stationTriplet'],
                    parameter=row['elementCd'],
                    start_date=self.getvalue('start_date'),
                    end_date=self.getvalue('end_date'),
                    debug=self.debug,
                )
            except NoData:
                continue

            data.append(row)

        self.data = data


class ElementIO(SnotelIO):
    """
    List of all SNOTEL element names, codes and units.
    """
    data_function = 'getElements'

# Singleton instance
elements = ElementIO()


class DailyDataIO(SnotelIO):
    """
    Wrapper for getData(), used internally by StationDailyDataIO
    """
    data_function = 'getData'

    # Applicable WebserviceLoader default options
    station = FilterOpt(required=True, url_param='stationTriplets')
    parameter = FilterOpt(required=True, url_param='elementCd')
    start_date = DateOpt(required=True, url_param='beginDate')
    end_date = DateOpt(required=True, url_param='endDate')

    # HeightDepth parameters don't seem to be necessary.

    default_params = {
        'ordinal': 1,
        'duration': 'DAILY',
        'getFlags': 'true',
        'alwaysReturnDailyFeb29': 'false',
    }

    def parse(self):
        data = self.data[0]
        if not data or 'values' not in data:
            raise NoData
        bd = data['beginDate']
        ed = data['endDate']
        dates = fill_date_range(bd, ed, date_format='%Y-%m-%d %H:%M:%S')
        vals = as_list(data['values'])
        flags = as_list(data['flags'])

        self.data = [{
            'date': date,
            'value': val,
            'flag': flag
        } for date, val, flag in zip(dates, vals, flags)]


class StationDailyDataIO(StationDataIO):
    """
    Requests all daily data for the specified station, optionally filtered by
    parameter.  The outer IO is the list of available parameters/elements, with
    each item in the list containing a nested IO with the actual data.

    Usage:

    params = StationDailyDataIO(
        station='302:OR:SNTL',
        start_date='2014-07-01',
        end_date='2014-07-31'
    )
    for param in params:
        print param.element_name
        for row in param.data:
            print "   ", row.date, row.value, param.storedunitcd

    """
    inner_io_class = DailyDataIO
    duration = "DAILY"


class RegionDailyDataIO(StationIO):
    """
    All-in-one IO for loading site metadata and daily data for a region (i.e. a
    state, county, or basin).  Internally calls:
      - getStations()
      - getStationMetadata()
      - getStationElements()
      - getData()

    The outer IO is a list of sites in the region - derived from StationIO, but
    with an extra "data" property on each station pointing to an inner time
    series IO for each site.  The inner IO is based on StationDailyDataIO but
    flattened to avoid multiple levels of nesting.  parameter is optional but
    recommended (otherwise all available data for all sites will be returned).

    Usage:

    sites = RegionDailyDataIO(
        basin='17060105',
        start_date='2014-07-01',
        end_date='2014-07-31',
        parameter='TAVG',
    )
    for site in sites:
        print site.name
        for row in site.data:
            print "   ", row.date, row.value, row.storedunitcd

    """

    nested = True

    # Applicable WebserviceLoader default options
    start_date = DateOpt(required=True)
    end_date = DateOpt(required=True)
    parameter = FilterOpt(url_param='elementCds')

    @property
    def params(self):
        params = super(RegionDailyDataIO, self).params
        # Start and end date are actually only used by inner io.
        del params['start_date']
        del params['end_date']
        return params

    def load(self):
        super(RegionDailyDataIO, self).load()
        for station in self.data:
            station['data'] = flattened(
                StationDailyDataIO,
                station=station['stationTriplet'],
                start_date=self.getvalue('start_date'),
                end_date=self.getvalue('end_date'),
                parameter=self.getvalue('parameter'),
                debug=self.debug,
            )


class HourlyDataIO(TimeSeriesMapper, SnotelIO):
    """
    Wrapper for getHourlyData(), used internally by StationHourlyDataIO
    """
    data_function = 'getHourlyData'

    # TimeSeriesMapper configuration
    date_formats = [
        '%Y-%m-%d %H:%M:%S',
        '%Y-%m-%d %H:%M'
    ]

    # Applicable WebserviceLoader default options
    station = FilterOpt(required=True, url_param='stationTriplets')
    parameter = FilterOpt(required=True, url_param='elementCd')
    start_date = DateOpt(required=True, url_param='beginDate')
    end_date = DateOpt(required=True, url_param='endDate')

    # Additional options
    begin_hour = FilterOpt(url_param='beginHour')
    end_hour = FilterOpt(url_param='endHour')

    # HeightDepth parameters don't seem to be necessary.

    default_params = {
        'ordinal': 1,
    }

    def load(self):
        super(HourlyDataIO, self).load()
        if self.data and 'values' in self.data[0]:
            self.data = as_list(self.data[0]['values'])
        else:
            raise NoData


class StationHourlyDataIO(StationDataIO):
    """
    Requests all hourly data for the specified station, optionally filtered by
    parameter.  The outer IO is the list of available parameters/elements, with
    each item in the list containing a nested IO with the actual data.

    Usage:

    params = StationHourlyDataIO(
        station='302:OR:SNTL',
        start_date='2014-07-01',
        end_date='2014-07-02',
    )
    for param in params:
        print param.element_name
        for row in param.data:
            print "   ", row.datetime, row.value, param.storedunitcd

    """
    inner_io_class = HourlyDataIO
    duration = "HOURLY"


class ForecastPeriodIO(SnotelIO):
    data_function = 'getForecastPeriods'


class ForecastDataIO(SnotelIO):
    data_function = 'getForecast'

    station = FilterOpt(required=True, url_param='stationTriplets')
    parameter = FilterOpt(required=True, url_param='elementCd')

    forecast_period = FilterOpt(required=True, url_param='forecastPeriod')
    publication_date = DateOpt(required=True, url_param='publicationDate')


class ForecastIO(SnotelIO):
    data_function = 'getForecasts'

    station = FilterOpt(required=True, url_param='stationTriplet')
    parameter = FilterOpt(required=True, url_param='elementCd')

    forecast_period = FilterOpt(required=True, url_param='forecastPeriod')



from wq.io import CsvParser, TimeSeriesMapper, TupleMapper, BaseIO
from datetime import datetime, date, timedelta
from collections import OrderedDict
from climata.base import WebserviceLoader, FilterOpt


class HydrometLoader(WebserviceLoader):
    """
    Shared options for Hydromet IO classes.
    """

    # start_date and end_date are the same as WebserviceLoader defaults

    # None of the default region filters will work
    state = FilterOpt(ignored=True)
    county = FilterOpt(ignored=True)
    basin = FilterOpt(ignored=True)

    # Instead, specify a four-letter Hydromet station code (one per request)
    station = FilterOpt(required=True)

    # Hydromet parameter codes are also required (multiple allowed)
    parameter = FilterOpt(required=True, multi=True)


class HydrometIO(HydrometLoader, CsvParser, TimeSeriesMapper, BaseIO):
    """
    Base class for retrieving Hydromet data from USBR.
    Use DailyDataIO or InstantDataIO instead, depending on your needs.
    """

    # URL components (see url() below)
    region = 'pn'  # USBR region: pn or gp
    website = 'www.usbr.gov'
    script = None

    # TimeSeriesMapper configuration
    date_formats = ['%m/%d/%Y %H:%M', '%m/%d/%Y']

    def clean_field_name(self, field):
        return field.replace(self.getvalue('station').upper(), "").strip()

    # NetLoader configuration
    @property
    def url(self):
        if self.script is None:
            raise NotImplementedError("Script must be defined!")
        return "http://%s/%s-bin/%s" % (
            self.website, self.region, self.script
        )

    @property
    def params(self):
        start = self.getvalue('start_date')
        end = self.getvalue('end_date')
        params, params_is_complex = self.getlist('parameter')
        pcodes = [
            "%s %s" % (self.getvalue('station'), param)
            for param in params
        ]

        # Note: The USBR Perl scripts are pretty quirky: a specific ordering of
        # URL parameters is important for proper function.
        return OrderedDict([
            ('parameter', ",".join(pcodes)),
            ('syer', start.year),
            ('smnth', start.month),
            ('sdy', start.day),
            ('eyer', end.year),
            ('emnth', end.month),
            ('edy', end.day),
            ('format', 2),
        ])

    # Help CsvParser find data
    def reader_class(self):
        cls = super(HydrometIO, self).reader_class()

        class Reader(cls):
            def choose_header(self, rows):
                for i, row in enumerate(rows):
                    if len(row) > 0 and row[0] == "BEGIN DATA":
                        return i + 1, rows[i + 1]
                return 0, ['header_not_found']
        return Reader

    # Cancel BaseIO iteration when END DATA is seen
    # FIXME: what about len()?
    def usable_item(self, item):
        item = super(HydrometIO, self).usable_item(item)
        if item[0] == "END DATA":
            return None
        return item


class DailyDataIO(HydrometIO):
    """
    Retrieves daily values for a single USBR Hydromet/Agrimet site.

    Usage:

    data = DailyDataIO(station='ACAO', parameter=['GD','QD'])
    for row in data:
        print row.date, row.gd, row.qd
    """

    script = "webarccsv.pl"
    key_fields = ['date']


class MultiStationDailyIO(HydrometLoader, TupleMapper, BaseIO):
    """
    Retrieves daily values for one or more USBR Hydromet/Agrimet sites.
    (Internally calls DailyDataIO for each site.)

    Usage:

    data = MultiStationDailyIO(station=['ACAO'], parameter=['GD','QD'])
    for s in data:
        print s.station
        for row in s.data:
            print row.date, row.gd, row.qd
    """

    nested = True

    station = FilterOpt(required=True, multi=True)

    # Customize load function with nested IOs
    def load(self):
        self.data = [{
            'station': station,
            'data': DailyDataIO(
                station=station,
                parameter=self.getvalue('parameter'),
                start_date=self.getvalue('start_date'),
                end_date=self.getvalue('end_date'),
                debug=self.debug,
            )
        } for station in self.getvalue('station')]

    def parse(self):
        pass


class InstantDataIO(HydrometIO):
    """
    Retrieves instant (e.g. 15-min) values for USBR Hydromet/Agrimet sites.

    Usage:

    data = InstantDataIO(station='ACAO', parameter=['GH','Q'])
    for row in data:
        print row.datetime, row.gh, row.q
    """

    script = "webdaycsv.pl"
    key_fields = ['datetime']


# PN also has an agrimet.pl that can return all data from the last 2 weeks of
# data for a station (no params need to be specified)
class AgrimetRecentIO(InstantDataIO):
    """
    Load recent Agrimet data (all available parameters)

    Usage:

    data = AgrimetRecentIO(station='ABEI')
    for row in data:
        print row.datetime, row.ob, row.wd
    """
    script = "agrimet.pl"

    start_date = FilterOpt(ignored=True)
    end_date = FilterOpt(ignored=True)
    parameter = FilterOpt(ignored=True)

    @property
    def params(self):
        return OrderedDict([
            ('cbtt', self.getvalue('station')),
            ('interval', 'instant'),
            ('format', 2),
            ('back', 360)
        ])



from wq.io import CsvNetIO
from collections import OrderedDict


class CountyIO(CsvNetIO):
    url = "http://www.census.gov/geo/reference/codes/files/national_county.txt"


counties = OrderedDict(sorted([
    (c.stateansi + c.countyansi, c)
    for c in CountyIO()
]))


def state_counties(state):
    result = OrderedDict()
    for fips, county in counties.items():
        if county.stateansi == state or county.state == state:
            result[fips] = county
    return result



DATA_TYPES = {
    'all': 'default',
    'iv': 'instantaneous Values',
    'uv': 'Unit Values',
    'rt': 'Real Time Values',
    'dv': 'Daily Values',
    'pk': 'Peak Measurements',
    'sv': 'Site Visits',
    'gw': 'Groundwater Levels',
    'qw': 'Water Quality Data',
    'id': 'Historical Instantaneous Values',
    'aw': 'Sites Monitored by the Active Groundwater Network',
    'ad': 'Sites included in Annual Water Data Reports',
}


SITE_TYPES = {
    'AG': 'Aggregate groundwater use',
    'AS': 'Aggregate surface-water-use',
    'AT': 'Atmosphere',
    'AW': 'Aggregate water-use establishment',
    'ES': 'Estuary',
    'FA': 'Facility',
    'FA-AWL': 'Animal waste lagoon',
    'FA-CI': 'Cistern',
    'FA-CS': 'Combined sewer',
    'FA-DV': 'Diversion',
    'FA-FON': 'Field, Pasture, Orchard, or Nursery',
    'FA-GC': 'Golf course',
    'FA-HP': 'Hydroelectric plant',
    'FA-LF': 'Landfill',
    'FA-OF': 'Outfall',
    'FA-PV': 'Pavement',
    'FA-QC': 'Laboratory or sample-preparation area',
    'FA-SEW': 'Wastewater sewer',
    'FA-SPS': 'Septic system',
    'FA-STS': 'Storm sewer',
    'FA-TEP': 'Thermoelectric plant',
    'FA-WDS': 'Water-distribution system',
    'FA-WIW': 'Waste injection well',
    'FA-WTP': 'Water-supply treatment plant',
    'FA-WU': 'Water-use establishment',
    'FA-WWD': 'Wastewater land application',
    'FA-WWTP': 'Wastewater-treatment plant',
    'GL': 'Glacier',
    'GW': 'Well',
    'GW-CR': 'Collector or Ranney type well',
    'GW-EX': 'Extensometer well',
    'GW-HZ': 'Hyporheic-zone well',
    'GW-IW': 'Interconnected wells',
    'GW-MW': 'Multiple wells',
    'GW-TH': 'Test hole not completed as a well',
    'LA': 'Land',
    'LA-EX': 'Excavation',
    'LA-OU': 'Outcrop',
    'LA-PLY': 'Playa',
    'LA-SH': 'Soil hole',
    'LA-SNK': 'Sinkhole',
    'LA-SR': 'Shore',
    'LA-VOL': 'Volcanic vent',
    'LK': 'Lake, Reservoir, Impoundment',
    'OC': 'Ocean',
    'OC-CO': 'Coastal',
    'SB': 'Subsurface',
    'SB-CV': 'Cave',
    'SB-GWD': 'Groundwater drain',
    'SB-TSM': 'Tunnel, shaft, or mine',
    'SB-UZ': 'Unsaturated zone',
    'SP': 'Spring',
    'SS': 'Specific Source',
    'ST': 'Stream',
    'ST-CA': 'Canal',
    'ST-DCH': 'Ditch',
    'ST-TS': 'Tidal stream',
    'WE': 'Wetland',
}



import json
from wq.io import JsonParser, BaseIO, TupleMapper, TimeSeriesMapper
from climata.base import (
    WebserviceLoader, FilterOpt, DateOpt, ChoiceOpt,
    parse_date, fill_date_range,
)
from .constants import *


class ParameterOpt(ChoiceOpt):
    multi = True
    url_param = 'elems'
    choices = ELEMENT_BY_ID.keys() + ELEMENT_BY_NAME.keys()


class AcisIO(WebserviceLoader, JsonParser, TupleMapper, BaseIO):
    """
    Base class for loading data from ACIS web services
    See http://data.rcc-acis.org/doc/
    """

    path = None  # ACIS web service path

    # (Re-)define some default WebserviceLoader options
    state = FilterOpt(multi=True)
    county = FilterOpt(multi=True)
    basin = FilterOpt(multi=True)
    station = FilterOpt(ignored=True)

    # Additional ACIS-specific option
    meta = ChoiceOpt(
        multi=True,
        choices=ALL_META_FIELDS,
        default=DEFAULT_META_FIELDS,
    )

    @property
    def url(self):
        """
        URL for wq.io.loaders.NetLoader
        """
        return "http://data.rcc-acis.org/%s" % self.path

    def serialize_params(self, params, complex):
        if complex:
            # ACIS web service supports JSON object as "params" parameter
            nparams = {}
            for key, val in params.items():
                url_param = self.get_url_param(key)
                if len(val) == 1 and isinstance(val[0], basestring):
                    val = val[0]
                nparams[url_param] = val
            return {'params': json.dumps(nparams)}
        else:
            # Simpler queries can use traditional URL parameters
            return super(AcisIO, self).serialize_params(params)


class StationMetaIO(AcisIO):
    """
    Retrieves metadata about the climate stations in a region.
    See http://data.rcc-acis.org/doc/#title8
    """

    namespace = "meta"  # For wq.io.parsers.text.JsonParser
    path = "StnMeta"

    # These options are not required for StationMetaIO
    start_date = DateOpt(url_param='sdate')
    end_date = DateOpt(url_param='edate')
    parameter = ParameterOpt()

    def parse(self):
        """
        Convert ACIS 'll' value into separate latitude and longitude.
        """
        super(AcisIO, self).parse()

        # This is more of a "mapping" step than a "parsing" step, but mappers
        # only allow one-to-one mapping from input fields to output fields.
        for row in self.data:
            if 'meta' in row:
                row = row['meta']
            if 'll' in row:
                row['longitude'], row['latitude'] = row['ll']
                del row['ll']

    def map_value(self, field, value):
        """
        Clean up some values returned from the web service.
        (overrides wq.io.mappers.BaseMapper)
        """

        if field == 'sids':
            # Site identifiers are returned as "[id] [auth_id]";
            # Map to auth name for easier usability
            ids = {}
            for idinfo in value:
                id, auth = idinfo.split(' ')
                auth = AUTHORITY_BY_ID[auth]
                ids[auth['name']] = id
            return ids

        if field == 'valid_daterange':
            # Date ranges for each element are returned in an array
            # (sorted by the order the elements were were requested);
            # Convert to dictionary with element id as key
            elems, complex = self.getlist('parameter')
            ranges = {}
            for elem, val in zip(elems, value):
                if val:
                    start, end = val
                    ranges[elem] = (parse_date(start), parse_date(end))
                else:
                    ranges[elem] = None, None
            return ranges
        return value


class StationDataIO(StationMetaIO):
    """
    Retrieve daily time series data from the climate stations in a region.
    See http://data.rcc-acis.org/doc/#title19
    """

    nested = True

    namespace = "data"  # For wq.io.parsers.text.JsonParser
    path = "MultiStnData"

    # Specify ACIS-defined URL parameters for start/end date
    start_date = DateOpt(required=True, url_param='sdate')
    end_date = DateOpt(required=True, url_param='edate')

    parameter = ParameterOpt(required=True)

    # Additional information for daily results
    add = ChoiceOpt(multi=True, choices=ADD_IDS)

    def get_field_names(self):
        """
        ACIS web service returns "meta" and "data" for each station;
        Use meta attributes as field names
        """
        field_names = super(StationDataIO, self).get_field_names()
        if field_names == ['meta', 'data']:
            meta_fields = self.data[0]['meta'].keys()
            if set(meta_fields) < set(self.getvalue('meta')):
                meta_fields = self.getvalue('meta')
            field_names = list(meta_fields) + ['data']
        return field_names

    def serialize_params(self, params, complex):
        # If set, apply "add" option to each requested element / parameter
        # (Rather than as a top-level URL param)
        if 'add' in params:
            complex = True
            elems = []
            for elem in params.get('parameter', []):
                if not isinstance(elem, dict):
                    elem = {'name': elem}
                elem['add'] = ",".join(params['add'])
                elems.append(elem)
            params['parameter'] = elems
            del params['add']
        return super(StationDataIO, self).serialize_params(params, complex)

    def usable_item(self, data):
        """
        ACIS web service returns "meta" and "data" for each station; use meta
        attributes as item values, and add an IO for iterating over "data"
        """

        # Use metadata as item
        item = data['meta']

        # Add nested IO for data
        elems, elems_is_complex = self.getlist('parameter')
        if elems_is_complex:
            elems = [elem['name'] for elem in elems]

        add, add_is_complex = self.getlist('add')
        item['data'] = DataIO(
            data=data['data'],
            parameter=elems,
            add=add,
            start_date=self.getvalue('start_date'),
            end_date=self.getvalue('end_date'),
        )

        # TupleMapper will convert item to namedtuple
        return super(StationDataIO, self).usable_item(item)


class DataIO(TimeSeriesMapper, BaseIO):
    """
    IO for iterating over ACIS time series data.
    Created internally by StationDataIO; not meant to be used directly.
    """

    # Inherited from parent
    parameter = []
    add = []
    start_date = None
    end_date = None

    date_formats = []  # For TimeSeriesMapper

    def load_data(self, data):
        """
        MultiStnData data results are arrays without explicit dates;
        Infer time series based on start date.
        """

        dates = fill_date_range(self.start_date, self.end_date)
        for row, date in zip(data, dates):
            data = {'date': date}
            if self.add:
                # If self.add is set, results will contain additional
                # attributes (e.g. flags). In that case, create one row per
                # result, with attributes "date", "elem", "value", and one for
                # each item in self.add.
                for elem, vals in zip(self.parameter, row):
                    data['elem'] = elem
                    for add, val in zip(['value'] + self.add, vals):
                        data[add] = val
                    yield data
            else:
                # Otherwise, return one row per date, with "date" and each
                # element's value as attributes.
                for elem, val in zip(self.parameter, row):
                    # namedtuple doesn't like numeric field names
                    if elem.isnumeric():
                        elem = "e%s" % elem
                    data[elem] = val
                yield data

    def __init__(self, *args, **kwargs):
        data = kwargs.pop('data')
        super(DataIO, self).__init__(*args, **kwargs)
        self.data = list(self.load_data(data))

    def get_field_names(self):
        """
        Different field names depending on self.add setting (see load_data)
        For BaseIO
        """
        if self.add:
            return ['date', 'elem', 'value'] + [flag for flag in self.add]
        else:
            field_names = ['date']
            for elem in self.parameter:
                # namedtuple doesn't like numeric field names
                if elem.isnumeric():
                    elem = "e%s" % elem
                field_names.append(elem)
            return field_names

    @property
    def key_fields(self):
        """
        Different key fields depending on self.add setting (see load_data)
        For TimeSeriesMapper
        """
        if self.add:
            return ['date', 'elem']
        else:
            return ['date']



from wq.io import CsvParser, BaseIO, TimeSeriesMapper
from wq.io.parsers.base import BaseParser
from owslib.waterml.wml11 import WaterML_1_1 as WaterML


class RdbParser(CsvParser):
    """
    Parser for RDB format (basically TSV with an extra header row)
    """
    max_header_row = 100
    delimiter = '\t'

    def parse(self):
        super(RdbParser, self).parse()
        # Skip second row which contains column lengths (5s, 10s, etc.)
        self.data = self.data[1:]


class TimeSeriesIO(TimeSeriesMapper, BaseIO):
    """
    Inner IO class for use by WaterMlParser.
    """

    # owslib already handles date parsing
    date_formats = []


class WaterMlParser(BaseParser):
    """
    wq.io-compatible Parser mixin for WaterML timeseries data.
    Generates a nested IO for each actual time series.

    Leverages owslib.waterml internally.
    """

    nested = True

    def parse(self):
        response = WaterML(self.file.read()).response
        self.response = response
        self.data = map(self.parse_timeseries, response.time_series)

    def parse_timeseries(self, ts):
        site = ts.source_info
        param = ts.variable
        lng, lat = map(float, site.location.geo_coords[0])

        # FIXME: This assumes there is only one values array, which might not
        # always be the case? (Same for site_codes below and geo_coords above.)
        values = ts.values[0]
        data = [{
            'date': date,
            'value': value,
        } for date, value in values.get_date_values()]

        return {
            'site_name': site.site_name,
            'site_code': site.site_codes[0],
            'variable_name': param.variable_name,
            'variable_code': param.variable_code,
            'unit': param.unit.code,
            'latitude': lat,
            'longitude': lng,
            'data': TimeSeriesIO(data=data)
        }



from wq.io import NetLoader, TupleMapper, TimeSeriesMapper, BaseIO
from climata.base import WebserviceLoader, FilterOpt, DateOpt, ChoiceOpt
from climata.parsers import RdbParser, WaterMlParser
from .constants import SITE_TYPES


# Base/mixin classes (not meant to be called directly)

class NwisLoader(WebserviceLoader):
    """
    Base class for loading data from the USGS NWIS REST services.
    """

    # Map WebserviceLoader options to NWIS equivalents
    start_date = DateOpt(url_param='startDT')
    end_date = DateOpt(url_param='endDT')

    state = FilterOpt(url_param='stateCd')
    county = FilterOpt(url_param='countyCd', multi=True)
    basin = FilterOpt(url_param='huc', multi=True)

    station = FilterOpt(url_param='site', multi=True)
    parameter = FilterOpt(url_param='parameterCd', multi=True)

    # Additional options unique to NWIS
    sitetype = ChoiceOpt(
        url_param='siteType',
        multi=True,
        choices=SITE_TYPES.keys(),
    )

    # Each NWIS webservice uses the same base URL, with a service path
    service = None

    @property
    def url(self):
        return "http://waterservices.usgs.gov/nwis/%s/" % self.service


class NwisWaterMlIO(NwisLoader, WaterMlParser, TupleMapper, BaseIO):
    """
    Base class for loading WaterML data from the USGS web services.  Use
    DailyValuesIO or InstantValuesIO instead depending on your needs.
    """

    default_params = {
        'format': 'waterml,1.1',
    }


class ParameterIO(NetLoader, RdbParser, TupleMapper, BaseIO):
    """
    Base class for loading USGS NWIS parameter code definitions.  Use
    FixedParameterIO or NumericParameterIO instead depending on your needs.
    """
    query = None
    params = {'fmt': 'rdb'}

    @property
    def url(self):
        return "http://help.waterdata.usgs.gov/code/%s" % self.query


# Preset classes for commonly used NWIS web services

class SiteIO(NwisLoader, RdbParser, TupleMapper, BaseIO):
    """
    Loads USGS site metadata from NWIS webservices (via RDB format).

    Usage:

        site_params = SiteIO(basin='02070010')
        for site_param in site_params:
            print site_param.site_no, site_param.parm_cd, site_param.end_date
    """

    service = "site"
    default_params = {
        'format': 'rdb,1.0',
        'outputDataTypeCd': 'all',
        'siteStatus': 'active',
    }


class DailyValueIO(NwisWaterMlIO):
    """
    Load USGS daily values from NWIS webservices (via WaterML format).

    Usage:

        dvals = DailyValueIO(basin='02070010')
        for dval in dvals:
            print
            print dval.site_code, dval.variable_code
            for row in dval.data:
                print row.date, row.value
    """

    service = 'dv'


class InstantValueIO(NwisWaterMlIO):
    """
    Load USGS instant values from NWIS webservices (via WaterML format).

    Usage:

        ivals = InstantValueIO(basin='02070010')
        for ival in ivals:
            print
            print ival.site_code, ival.variable_code
            for row in ival.data:
                print row.date, row.value
    """

    service = 'iv'


# Preset classes for USGS waterdata code definitions

class FixedParameterIO(ParameterIO):
    """
    Loads USGS fixed (text) parameters and value definitions.
    """

    query = "fixed_parms_query"


class NumericParameterIO(ParameterIO):
    """
    Loads USGS numeric parameter definitions.
    """

    query = "parameter_cd_query"
    params = {'fmt': 'rdb', 'group_cd': '%'}



from os.path import join, dirname
from setuptools import setup, find_packages


SERVICES = (
    "ACIS (NOAA RCCs)",
    "CoCoRaHS",
    "Hydromet (USBR)",
    "SNOTEL AWDB (NRCS)",
    "NWIS (USGS)",
)
SVC_STR = ", ".join(SERVICES[:-1]) + " and " + SERVICES[-1]

DESCRIPTION = "Loads climate and hydrology data from %s." % SVC_STR

LONG_DESCRIPTION = """
A pythonic library for iterating over climate time series data from %s.
Powered by wq.io.
""" % SVC_STR


def long_description():
    """Return long description from README.rst if it's present
    because it doesn't get installed."""
    try:
        return open(join(dirname(__file__), 'README.rst')).read()
    except IOError:
        return LONG_DESCRIPTION


def get_version():
    version = open(join(dirname(__file__), "climata", "version.py")).read()
    version = version.strip().replace("VERSION = ", '')
    version = version.replace('"', '')
    return version

setup(
    name='climata',
    version=get_version(),
    author='HEI Geo',
    author_email='gis@houstoneng.com',
    url='https://github.com/heigeo/climata',
    license='MIT',
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=long_description(),
    install_requires=['wq.io>=0.5.1', 'owslib', 'SOAPpy'],
    scripts=['climata/bin/acis_sites.py', 'climata/bin/acis_data.py'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: System :: Networking :: Monitoring',
    ]
)



from datetime import datetime
from warnings import warn
from wq.io import make_date_mapper, NetLoader
parse_date = make_date_mapper('%Y-%m-%d')

from .version import VERSION
from wq.io.loaders import HTTPLIB_VERSION, VERSION as WQIO_VERSION

from datetime import datetime, timedelta


class FilterOpt(object):
    """
    Base class for describing a filter option
    """
    name = None  # Option name as defined on IO class
    url_param = None  # Actual URL parameter to use (defaults to name)

    required = False  # Whether option is equired for valid request
    multi = False  # Whether multiple values are allowed
    ignored = False  # Used on subclasses when option does not apply

    default = None  # Default value

    def __init__(self, **kwargs):
        """
        Allow setting the above via kwargs
        """
        self.__dict__.update(**kwargs)

    def get_url_param(self):
        return self.url_param or self.name

    def parse(self, value):
        """
        Enforce rules and return parsed value
        """
        if self.required and value is None:
            raise ValueError("%s is required!" % self.name)
        elif self.ignored and value is not None:
            warn("%s is ignored for this class!" % self.name)
        elif not self.multi and isinstance(value, (list, tuple)):
            if len(value) > 1:
                raise ValueError(
                    "%s does not accept multiple values!" % self.name
                )
            return value[0]
        elif self.multi and value is not None:
            if not isinstance(value, (list, tuple)):
                return [value]
        return value


class DateOpt(FilterOpt):
    date_only = True

    def parse_date(self, value):
        return parse_date(value)

    def parse(self, value):
        """
        Parse date
        """
        value = super(DateOpt, self).parse(value)
        if value is None:
            return None
        if isinstance(value, basestring):
            value = self.parse_date(value)
        if isinstance(value, datetime) and self.date_only:
            value = value.date()
        return value


class ChoiceOpt(FilterOpt):
    choices = []  # Valid choices for this option

    def parse(self, value):
        value = super(ChoiceOpt, self).parse(value)
        if value is None:
            return None
        if isinstance(value, (list, tuple)):
            check_values = value
        else:
            check_values = [value]
        for cv in check_values:
            if cv not in self.choices:
                raise ValueError(
                    "%s is not a valid choice for %s!" % (cv, self.name)
                )
        return value


class WebserviceLoader(NetLoader):
    """
    NetLoader subclass with enhanced functionality for enumerating and
    validating URL arguments.
    """

    # Default filter options, common to most web services for climate data.
    # Every climata IO class is assumed to support at least these options; if
    # any are not applicable they should be overridden with ignored=True.

    # Time period filters
    start_date = DateOpt(required=True)
    end_date = DateOpt(required=True)

    # Region filters
    state = FilterOpt(multi=True)
    county = FilterOpt(multi=True)
    basin = FilterOpt(multi=True)

    # Station and parameter code filters
    station = FilterOpt(multi=True)
    parameter = FilterOpt(multi=True)

    # URL params that apply to every request (if any)
    default_params = {}

    def __init__(self, *args, **kwargs):
        """
        Initialize web service (and general IO) options
        """

        #  Validate web service parameters using FilterOpt information
        self._values = {}
        for name, opt in self.filter_options.items():
            opt.name = name
            val = kwargs.pop(name, opt.default)
            self._values[name] = opt.parse(val)

        # Mimic BaseIO behavior since it's not a super class of NetLoader
        if kwargs:
            self.__dict__.update(**kwargs)
        self.refresh()

    @classmethod
    def get_filter_options(cls):
        """
        List all filter options defined on class (and superclasses)
        """
        attr = '_filter_options_%s' % id(cls)

        options = getattr(cls, attr, {})
        if options:
            return options

        for key in dir(cls):
            val = getattr(cls, key)
            if isinstance(val, FilterOpt):
                options[key] = val

        setattr(cls, attr, options)
        return options

    @property
    def filter_options(self):
        return type(self).get_filter_options()

    def get_url_param(self, key):
        return self.filter_options[key].get_url_param()

    def getvalue(self, name):
        return self._values[name]

    def getlist(self, name):
        """
        Retrieve given property from class/instance, ensuring it is a list.
        Also determine whether the list contains simple text/numeric values or
        nested dictionaries (a "complex" list)
        """
        value = self.getvalue(name)
        complex = False

        def str_value(val):
            if isinstance(val, dict):
                complex = True
                return val
            else:
                return unicode(val)

        if value is None:
            pass
        else:
            value = [str_value(val) for val in as_list(value)]

        return value, complex

    def set_param(self, into, name):
        """
        Set parameter key, noting whether list value is "complex"
        """
        value, complex = self.getlist(name)
        if value is not None:
            into[name] = value
        return complex

    def get_params(self):
        """
        Get parameters for web service, noting whether any are "complex"
        """
        params = {}
        complex = False

        for name, opt in self.filter_options.items():
            if opt.ignored:
                continue
            if self.set_param(params, name):
                complex = True
        return params, complex

    @property
    def params(self):
        """
        URL parameters for wq.io.loaders.NetLoader
        """
        params, complex = self.get_params()
        url_params = self.default_params.copy()
        url_params.update(self.serialize_params(params, complex))
        return url_params

    def serialize_params(self, params, complex=False):
        """
        Serialize parameter names and values to a dict ready for urlencode()
        """
        if complex:
            # See climata.acis for an example implementation
            raise NotImplementedError("Cannot serialize %s!" % params)
        else:
            # Simpler queries can use traditional URL parameters
            return {
                self.get_url_param(key): ','.join(val)
                for key, val in params.items()
            }

    @property
    def user_agent(self):
        agent = "climata %s (wq.io/%s; Python-httplib2/%s) %s" % (
            VERSION, WQIO_VERSION, HTTPLIB_VERSION,
            "https://github.com/heigeo/climata"
        )
        return agent


def fill_date_range(start_date, end_date, date_format=None):
    """
    Function accepts start date, end date, and format (if dates are strings)
    and returns a list of Python dates.
    """

    if date_format:
        start_date = datetime.strptime(start_date, date_format).date()
        end_date = datetime.strptime(end_date, date_format).date()
    date_list = []
    while start_date <= end_date:
        date_list.append(start_date)
        start_date = start_date + timedelta(days=1)
    return date_list


def as_list(value):
    if isinstance(value, (list, tuple)):
        return value
    else:
        return [value]
